# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("d3d10.dll")
prototypes = \
    {
        # 
        'D3D10CreateDevice': SimTypeFunction([SimTypeBottom(label="IDXGIAdapter"), SimTypeInt(signed=False, label="D3D10_DRIVER_TYPE"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3D10Device"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAdapter", "DriverType", "Software", "Flags", "SDKVersion", "ppDevice"]),
        # 
        'D3D10CreateDeviceAndSwapChain': SimTypeFunction([SimTypeBottom(label="IDXGIAdapter"), SimTypeInt(signed=False, label="D3D10_DRIVER_TYPE"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"BufferDesc": SimStruct({"Width": SimTypeInt(signed=False, label="UInt32"), "Height": SimTypeInt(signed=False, label="UInt32"), "RefreshRate": SimStruct({"Numerator": SimTypeInt(signed=False, label="UInt32"), "Denominator": SimTypeInt(signed=False, label="UInt32")}, name="DXGI_RATIONAL", pack=False, align=None), "Format": SimTypeInt(signed=False, label="DXGI_FORMAT"), "ScanlineOrdering": SimTypeInt(signed=False, label="DXGI_MODE_SCANLINE_ORDER"), "Scaling": SimTypeInt(signed=False, label="DXGI_MODE_SCALING")}, name="DXGI_MODE_DESC", pack=False, align=None), "SampleDesc": SimStruct({"Count": SimTypeInt(signed=False, label="UInt32"), "Quality": SimTypeInt(signed=False, label="UInt32")}, name="DXGI_SAMPLE_DESC", pack=False, align=None), "BufferUsage": SimTypeInt(signed=False, label="UInt32"), "BufferCount": SimTypeInt(signed=False, label="UInt32"), "OutputWindow": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Windowed": SimTypeInt(signed=True, label="Int32"), "SwapEffect": SimTypeInt(signed=False, label="DXGI_SWAP_EFFECT"), "Flags": SimTypeInt(signed=False, label="UInt32")}, name="DXGI_SWAP_CHAIN_DESC", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="IDXGISwapChain"), offset=0), SimTypePointer(SimTypeBottom(label="ID3D10Device"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAdapter", "DriverType", "Software", "Flags", "SDKVersion", "pSwapChainDesc", "ppSwapChain", "ppDevice"]),
        # 
        'D3D10CreateBlob': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["NumBytes", "ppBuffer"]),
        # 
        'D3D10CompileShader': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"Name": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Definition": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="D3D_SHADER_MACRO", pack=False, align=None), offset=0), SimTypeBottom(label="ID3DInclude"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "pFileName", "pDefines", "pInclude", "pFunctionName", "pProfile", "Flags", "ppShader", "ppErrorMsgs"]),
        # 
        'D3D10DisassembleShader': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pShader", "BytecodeLength", "EnableColorCode", "pComments", "ppDisassembly"]),
        # 
        'D3D10GetPixelShaderProfile': SimTypeFunction([SimTypeBottom(label="ID3D10Device")], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["pDevice"]),
        # 
        'D3D10GetVertexShaderProfile': SimTypeFunction([SimTypeBottom(label="ID3D10Device")], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["pDevice"]),
        # 
        'D3D10GetGeometryShaderProfile': SimTypeFunction([SimTypeBottom(label="ID3D10Device")], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["pDevice"]),
        # 
        'D3D10ReflectShader': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3D10ShaderReflection"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pShaderBytecode", "BytecodeLength", "ppReflector"]),
        # 
        'D3D10PreprocessShader': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"Name": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Definition": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="D3D_SHADER_MACRO", pack=False, align=None), offset=0), SimTypeBottom(label="ID3DInclude"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "pFileName", "pDefines", "pInclude", "ppShaderText", "ppErrorMsgs"]),
        # 
        'D3D10GetInputSignatureBlob': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pShaderBytecode", "BytecodeLength", "ppSignatureBlob"]),
        # 
        'D3D10GetOutputSignatureBlob': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pShaderBytecode", "BytecodeLength", "ppSignatureBlob"]),
        # 
        'D3D10GetInputAndOutputSignatureBlob': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pShaderBytecode", "BytecodeLength", "ppSignatureBlob"]),
        # 
        'D3D10GetShaderDebugInfo': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pShaderBytecode", "BytecodeLength", "ppDebugInfo"]),
        # 
        'D3D10StateBlockMaskUnion': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pA", "pB", "pResult"]),
        # 
        'D3D10StateBlockMaskIntersect': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pA", "pB", "pResult"]),
        # 
        'D3D10StateBlockMaskDifference': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pA", "pB", "pResult"]),
        # 
        'D3D10StateBlockMaskEnableCapture': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="D3D10_DEVICE_STATE_TYPES"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pMask", "StateType", "RangeStart", "RangeLength"]),
        # 
        'D3D10StateBlockMaskDisableCapture': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="D3D10_DEVICE_STATE_TYPES"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pMask", "StateType", "RangeStart", "RangeLength"]),
        # 
        'D3D10StateBlockMaskEnableAll': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pMask"]),
        # 
        'D3D10StateBlockMaskDisableAll': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pMask"]),
        # 
        'D3D10StateBlockMaskGetSetting': SimTypeFunction([SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="D3D10_DEVICE_STATE_TYPES"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pMask", "StateType", "Entry"]),
        # 
        'D3D10CreateStateBlock': SimTypeFunction([SimTypeBottom(label="ID3D10Device"), SimTypePointer(SimStruct({"VS": SimTypeChar(label="Byte"), "VSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "VSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "VSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GS": SimTypeChar(label="Byte"), "GSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "GSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "GSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PS": SimTypeChar(label="Byte"), "PSSamplers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "PSShaderResources": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "PSConstantBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAVertexBuffers": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "IAIndexBuffer": SimTypeChar(label="Byte"), "IAInputLayout": SimTypeChar(label="Byte"), "IAPrimitiveTopology": SimTypeChar(label="Byte"), "OMRenderTargets": SimTypeChar(label="Byte"), "OMDepthStencilState": SimTypeChar(label="Byte"), "OMBlendState": SimTypeChar(label="Byte"), "RSViewports": SimTypeChar(label="Byte"), "RSScissorRects": SimTypeChar(label="Byte"), "RSRasterizerState": SimTypeChar(label="Byte"), "SOBuffers": SimTypeChar(label="Byte"), "Predication": SimTypeChar(label="Byte")}, name="D3D10_STATE_BLOCK_MASK", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3D10StateBlock"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDevice", "pStateBlockMask", "ppStateBlock"]),
        # 
        'D3D10CompileEffectFromMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"Name": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Definition": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="D3D_SHADER_MACRO", pack=False, align=None), offset=0), SimTypeBottom(label="ID3DInclude"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pData", "DataLength", "pSrcFileName", "pDefines", "pInclude", "HLSLFlags", "FXFlags", "ppCompiledEffect", "ppErrors"]),
        # 
        'D3D10CreateEffectFromMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="ID3D10Device"), SimTypeBottom(label="ID3D10EffectPool"), SimTypePointer(SimTypeBottom(label="ID3D10Effect"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pData", "DataLength", "FXFlags", "pDevice", "pEffectPool", "ppEffect"]),
        # 
        'D3D10CreateEffectPoolFromMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="ID3D10Device"), SimTypePointer(SimTypeBottom(label="ID3D10EffectPool"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pData", "DataLength", "FXFlags", "pDevice", "ppEffectPool"]),
        # 
        'D3D10DisassembleEffect': SimTypeFunction([SimTypeBottom(label="ID3D10Effect"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pEffect", "EnableColorCode", "ppDisassembly"]),
    }

lib.set_prototypes(prototypes)



import angr

import itertools
rand_count = itertools.count()

class random(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, buf, count, rnd_bytes):
        if self.state.mode == 'fastpath':
            # Special case for CFG
            if (not self.state.solver.symbolic(count) and
                    not self.state.solver.symbolic(buf) and
                    not self.state.solver.symbolic(rnd_bytes)):
                if (self.state.solver.is_true(rnd_bytes != 0) and self.state.cgc.addr_invalid(rnd_bytes))\
                        or self.state.cgc.addr_invalid(buf):
                    return self.state.cgc.EFAULT

                max_count = self.state.solver.eval_one(count)
                random_num = self.state.solver.Unconstrained('random_%d' % next(rand_count), max_count * 8)
                self.state.memory.store(buf, random_num, size=count)
                if self.state.solver.is_true(rnd_bytes != 0):
                    self.state.memory.store(rnd_bytes, count, endness='Iend_LE')

            # We always return something in fastpath mode
            return self.state.solver.BVV(0, self.state.arch.bits)

        # return code
        r = self.state.solver.ite_cases((
                (self.state.cgc.addr_invalid(buf), self.state.cgc.EFAULT),
                (self.state.solver.And(rnd_bytes != 0, self.state.cgc.addr_invalid(rnd_bytes)), self.state.cgc.EFAULT),
            ), self.state.solver.BVV(0, self.state.arch.bits))

        if self.state.satisfiable(extra_constraints=[count!=0]):
            max_size = min(
                    self.state.solver.max_int(count * 8),
                    0x10000,
                    )
            self.state.memory.store(buf, self.state.solver.Unconstrained('random_%d' % next(rand_count), max_size, key=('syscall', 'random')), size=count)
        self.state.memory.store(rnd_bytes, count, endness='Iend_LE', condition=rnd_bytes != 0)

        return r



from ailment.expression import BinaryOp, Const, UnaryOp, Convert

from .base import PeepholeOptimizationExprBase


class BoolExprXor1(PeepholeOptimizationExprBase):
    __slots__ = ()

    name = "bool_expr ^ 1 => !bool_expr (a)"
    expr_classes = (BinaryOp, )  # all expressions are allowed

    def optimize(self, expr: BinaryOp):

        # Conv(1->N, some_bool_expr) ^ 1 ==> Conv(1->N, Not(some_bool_expr))
        if expr.op == "Xor" and isinstance(expr.operands[1], Const) and expr.operands[1].value == 1:
            arg0 = expr.operands[0]
            if isinstance(arg0, Convert) and arg0.from_bits == 1 \
                    and self.is_bool_expr(arg0.operand):
                new_expr = Convert(None, 1, arg0.to_bits, arg0.is_signed,
                                   UnaryOp(None, 'Not', arg0.operands[0], **expr.tags),
                                   **arg0.tags)
                return new_expr

        return None



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("usp10.dll")
prototypes = \
    {
        # 
        'ScriptFreeCache': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["psc"]),
        # 
        'ScriptItemize': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="SCRIPT_CONTROL", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"iCharPos": SimTypeInt(signed=True, label="Int32"), "a": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None)}, name="SCRIPT_ITEM", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwcInChars", "cInChars", "cMaxItems", "psControl", "psState", "pItems", "pcItems"]),
        # 
        'ScriptLayout': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cRuns", "pbLevel", "piVisualToLogical", "piLogicalToVisual"]),
        # 
        'ScriptShape': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "pwcChars", "cChars", "cMaxGlyphs", "psa", "pwOutGlyphs", "pwLogClust", "psva", "pcGlyphs"]),
        # 
        'ScriptPlace': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimStruct({"du": SimTypeInt(signed=True, label="Int32"), "dv": SimTypeInt(signed=True, label="Int32")}, name="GOFFSET", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"abcA": SimTypeInt(signed=True, label="Int32"), "abcB": SimTypeInt(signed=False, label="UInt32"), "abcC": SimTypeInt(signed=True, label="Int32")}, name="ABC", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "pwGlyphs", "cGlyphs", "psva", "psa", "piAdvance", "pGoffset", "pABC"]),
        # 
        'ScriptTextOut': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimStruct({"du": SimTypeInt(signed=True, label="Int32"), "dv": SimTypeInt(signed=True, label="Int32")}, name="GOFFSET", pack=False, align=None), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "x", "y", "fuOptions", "lprc", "psa", "pwcReserved", "iReserved", "pwGlyphs", "cGlyphs", "piAdvance", "piJustify", "pGoffset"]),
        # 
        'ScriptJustify': SimTypeFunction([SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["psva", "piAdvance", "cGlyphs", "iDx", "iMinKashida", "piJustify"]),
        # 
        'ScriptBreak': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeChar(label="Byte")}, name="SCRIPT_LOGATTR", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwcChars", "cChars", "psa", "psla"]),
        # 
        'ScriptCPtoX': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["iCP", "fTrailing", "cChars", "cGlyphs", "pwLogClust", "psva", "piAdvance", "psa", "piX"]),
        # 
        'ScriptXtoCP': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["iX", "cChars", "cGlyphs", "pwLogClust", "psva", "piAdvance", "psa", "piCP", "piTrailing"]),
        # 
        'ScriptGetLogicalWidths': SimTypeFunction([SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["psa", "cChars", "cGlyphs", "piGlyphWidth", "pwLogClust", "psva", "piDx"]),
        # 
        'ScriptApplyLogicalWidth': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"abcA": SimTypeInt(signed=True, label="Int32"), "abcB": SimTypeInt(signed=False, label="UInt32"), "abcC": SimTypeInt(signed=True, label="Int32")}, name="ABC", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["piDx", "cChars", "cGlyphs", "pwLogClust", "psva", "piAdvance", "psa", "pABC", "piJustify"]),
        # 
        'ScriptGetCMap': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "pwcInChars", "cChars", "dwFlags", "pwOutGlyphs"]),
        # 
        'ScriptGetGlyphABCWidth': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimStruct({"abcA": SimTypeInt(signed=True, label="Int32"), "abcB": SimTypeInt(signed=False, label="UInt32"), "abcC": SimTypeInt(signed=True, label="Int32")}, name="ABC", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "wGlyph", "pABC"]),
        # 
        'ScriptGetProperties': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypePointer(SimStruct({"_bitfield1": SimTypeInt(signed=False, label="UInt32"), "_bitfield2": SimTypeInt(signed=False, label="UInt32")}, name="SCRIPT_PROPERTIES", pack=False, align=None), offset=0), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppSp", "piNumScripts"]),
        # 
        'ScriptGetFontProperties': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"cBytes": SimTypeInt(signed=True, label="Int32"), "wgBlank": SimTypeShort(signed=False, label="UInt16"), "wgDefault": SimTypeShort(signed=False, label="UInt16"), "wgInvalid": SimTypeShort(signed=False, label="UInt16"), "wgKashida": SimTypeShort(signed=False, label="UInt16"), "iKashidaWidth": SimTypeInt(signed=True, label="Int32")}, name="SCRIPT_FONTPROPERTIES", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "sfp"]),
        # 
        'ScriptCacheGetHeight': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "tmHeight"]),
        # 
        'ScriptStringAnalyse': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="SCRIPT_CONTROL", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimStruct({"cTabStops": SimTypeInt(signed=True, label="Int32"), "iScale": SimTypeInt(signed=True, label="Int32"), "pTabStops": SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), "iTabOrigin": SimTypeInt(signed=True, label="Int32")}, name="SCRIPT_TABDEF", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "pString", "cString", "cGlyphs", "iCharset", "dwFlags", "iReqWidth", "psControl", "psState", "piDx", "pTabdef", "pbInClass", "pssa"]),
        # 
        'ScriptStringFree': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pssa"]),
        # 
        'ScriptString_pSize': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypePointer(SimStruct({"cx": SimTypeInt(signed=True, label="Int32"), "cy": SimTypeInt(signed=True, label="Int32")}, name="SIZE", pack=False, align=None), offset=0), arg_names=["ssa"]),
        # 
        'ScriptString_pcOutChars': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), arg_names=["ssa"]),
        # 
        'ScriptString_pLogAttr': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypePointer(SimStruct({"_bitfield": SimTypeChar(label="Byte")}, name="SCRIPT_LOGATTR", pack=False, align=None), offset=0), arg_names=["ssa"]),
        # 
        'ScriptStringGetOrder': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ssa", "puOrder"]),
        # 
        'ScriptStringCPtoX': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ssa", "icp", "fTrailing", "pX"]),
        # 
        'ScriptStringXtoCP': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ssa", "iX", "piCh", "piTrailing"]),
        # 
        'ScriptStringGetLogicalWidths': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ssa", "piDx"]),
        # 
        'ScriptStringValidate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ssa"]),
        # 
        'ScriptStringOut': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="ETO_OPTIONS"), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["ssa", "iX", "iY", "uOptions", "prc", "iMinSel", "iMaxSel", "fDisabled"]),
        # 
        'ScriptIsComplex': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="SCRIPT_IS_COMPLEX_FLAGS")], SimTypeInt(signed=True, label="Int32"), arg_names=["pwcInChars", "cInChars", "dwFlags"]),
        # 
        'ScriptRecordDigitSubstitution': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"_bitfield1": SimTypeInt(signed=False, label="UInt32"), "_bitfield2": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32")}, name="SCRIPT_DIGITSUBSTITUTE", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Locale", "psds"]),
        # 
        'ScriptApplyDigitSubstitution': SimTypeFunction([SimTypePointer(SimStruct({"_bitfield1": SimTypeInt(signed=False, label="UInt32"), "_bitfield2": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32")}, name="SCRIPT_DIGITSUBSTITUTE", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="SCRIPT_CONTROL", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["psds", "psc", "pss"]),
        # 
        'ScriptShapeOpenType': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimStruct({"potfRecords": SimTypePointer(SimStruct({"tagFeature": SimTypeInt(signed=False, label="UInt32"), "lParameter": SimTypeInt(signed=True, label="Int32")}, name="opentype_feature_record", pack=False, align=None), offset=0), "cotfRecords": SimTypeInt(signed=True, label="Int32")}, name="textrange_properties", pack=False, align=None), offset=0), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="script_charprop", pack=False, align=None), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimStruct({"sva": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), "reserved": SimTypeShort(signed=False, label="UInt16")}, name="script_glyphprop", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "tagScript", "tagLangSys", "rcRangeChars", "rpRangeProperties", "cRanges", "pwcChars", "cChars", "cMaxGlyphs", "pwLogClust", "pCharProps", "pwOutGlyphs", "pOutGlyphProps", "pcGlyphs"]),
        # 
        'ScriptPlaceOpenType': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimStruct({"potfRecords": SimTypePointer(SimStruct({"tagFeature": SimTypeInt(signed=False, label="UInt32"), "lParameter": SimTypeInt(signed=True, label="Int32")}, name="opentype_feature_record", pack=False, align=None), offset=0), "cotfRecords": SimTypeInt(signed=True, label="Int32")}, name="textrange_properties", pack=False, align=None), offset=0), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="script_charprop", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimStruct({"sva": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_VISATTR", pack=False, align=None), "reserved": SimTypeShort(signed=False, label="UInt16")}, name="script_glyphprop", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimStruct({"du": SimTypeInt(signed=True, label="Int32"), "dv": SimTypeInt(signed=True, label="Int32")}, name="GOFFSET", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"abcA": SimTypeInt(signed=True, label="Int32"), "abcB": SimTypeInt(signed=False, label="UInt32"), "abcC": SimTypeInt(signed=True, label="Int32")}, name="ABC", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "tagScript", "tagLangSys", "rcRangeChars", "rpRangeProperties", "cRanges", "pwcChars", "pwLogClust", "pCharProps", "cChars", "pwGlyphs", "pGlyphProps", "cGlyphs", "piAdvance", "pGoffset", "pABC"]),
        # 
        'ScriptItemizeOpenType': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="SCRIPT_CONTROL", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"iCharPos": SimTypeInt(signed=True, label="Int32"), "a": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None)}, name="SCRIPT_ITEM", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwcInChars", "cInChars", "cMaxItems", "psControl", "psState", "pItems", "pScriptTags", "pcItems"]),
        # 
        'ScriptGetFontScriptTags': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "cMaxTags", "pScriptTags", "pcTags"]),
        # 
        'ScriptGetFontLanguageTags': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "tagScript", "cMaxTags", "pLangsysTags", "pcTags"]),
        # 
        'ScriptGetFontFeatureTags': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "tagScript", "tagLangSys", "cMaxTags", "pFeatureTags", "pcTags"]),
        # 
        'ScriptGetFontAlternateGlyphs': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "tagScript", "tagLangSys", "tagFeature", "wGlyphId", "cMaxAlternates", "pAlternateGlyphs", "pcAlternates"]),
        # 
        'ScriptSubstituteSingleGlyph': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "tagScript", "tagLangSys", "tagFeature", "lParameter", "wGlyphId", "pwOutGlyphId"]),
        # 
        'ScriptPositionSingleGlyph': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16"), "s": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="SCRIPT_STATE", pack=False, align=None)}, name="SCRIPT_ANALYSIS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeShort(signed=False, label="UInt16"), SimTypeInt(signed=True, label="Int32"), SimStruct({"du": SimTypeInt(signed=True, label="Int32"), "dv": SimTypeInt(signed=True, label="Int32")}, name="GOFFSET", pack=False, align=None), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimStruct({"du": SimTypeInt(signed=True, label="Int32"), "dv": SimTypeInt(signed=True, label="Int32")}, name="GOFFSET", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "psc", "psa", "tagScript", "tagLangSys", "tagFeature", "lParameter", "wGlyphId", "iAdvance", "GOffset", "piOutAdvance", "pOutGoffset"]),
    }

lib.set_prototypes(prototypes)



import cffi
from typing import Tuple, Type, Dict, Optional, Iterable, Set, Any
import logging
from collections import defaultdict

import claripy

from angr.storage.memory_mixins import MemoryMixin
from angr.storage.memory_mixins.paged_memory.pages import PageType, ListPage, UltraPage, MVListPage
from angr.storage.memory_object import SimLabeledMemoryObject
from ....errors import SimMemoryError

# yeet
ffi = cffi.FFI()

l = logging.getLogger(__name__)


class PagedMemoryMixin(MemoryMixin):
    """
    A bottom-level storage mechanism. Dispatches reads to individual pages, the type of which is the PAGE_TYPE class
    variable.
    """
    SUPPORTS_CONCRETE_LOAD = True
    PAGE_TYPE: Type[PageType] = None  # must be provided in subclass

    def __init__(self,  page_size=0x1000, default_permissions=3, permissions_map=None, page_kwargs=None, **kwargs):
        super().__init__(**kwargs)
        self.page_size = page_size
        self._extra_page_kwargs = page_kwargs if page_kwargs is not None else {}

        self._permissions_map = permissions_map if permissions_map is not None else {}
        self._default_permissions = default_permissions
        self._pages: Dict[int, Optional[PageType]] = {}

    @MemoryMixin.memo
    def copy(self, memo):
        o = super().copy(memo)

        o.page_size = self.page_size
        o._extra_page_kwargs = self._extra_page_kwargs

        o._default_permissions = self._default_permissions
        o._permissions_map = self._permissions_map
        o._pages = dict(self._pages)

        for page in o._pages.values():
            if page is not None:
                page.acquire_shared()

        return o

    def __del__(self):
        # a thought: we could support mapping pages in multiple places in memory if here we just kept a set of released
        # page ids and never released any page more than once
        for page in self._pages.values():
            if page is not None:
                page.release_shared()

    def _get_page(self, pageno: int, writing: bool, **kwargs) -> PageType:
        force_default = True
        # force_default means don't consult any "backers"
        # if None is stored explicitly in _pages, it means it was unmapped explicitly, so don't consult backers
        try:
            page = self._pages[pageno]
        except KeyError:
            page = None
            force_default = False

        if page is None:
            page = self._initialize_page(pageno, force_default=force_default, **kwargs)
            self._pages[pageno] = page

        if writing:
            page = page.acquire_unique()
            self._pages[pageno] = page
        return page

    def _initialize_default_page(self, pageno: int, permissions=None, **kwargs) -> PageType:
        # the difference between _initialize_default_page and _initialize_page with force_default=True
        # is that the latter may segfault. this strictly gives you a new page.
        kwargs['allow_default'] = True
        return PagedMemoryMixin._initialize_page(self, pageno, permissions=permissions, **kwargs)

    def _initialize_page(self, pageno: int, permissions=None, allow_default=True, **kwargs) -> PageType:
        if not allow_default:
            raise SimMemoryError("I have been instructed not to create a default page")

        return self.PAGE_TYPE(**self._page_kwargs(pageno, permissions))  # pylint:disable=not-callable

    def _page_kwargs(self, pageno, permissions):
        # permissions lookup: let permissions arg override everything else
        # then try the permissions map
        # then fall back to the default permissions
        if permissions is None:
            permissions = self._default_permissions
            addr = pageno * self.page_size

            for (start, end), perms in self._permissions_map.items():
                if start <= addr <= end:
                    permissions = perms
                    break

        return dict(
            memory=self,
            memory_id='%s_%d' % (self.id, pageno),
            permissions=permissions,
            **self._extra_page_kwargs
        )

    def _divide_addr(self, addr: int) -> Tuple[int, int]:
        return divmod(addr, self.page_size)

    def load(self, addr: int, size: int=None, endness=None, **kwargs):
        if endness is None:
            endness = self.endness

        if type(size) is not int:
            raise TypeError("Need size to be resolved to an int by this point")

        if type(addr) is not int:
            raise TypeError("Need addr to be resolved to an int by this point")

        pageno, pageoff = self._divide_addr(addr)
        vals = []

        # fasttrack basic case
        if pageoff + size <= self.page_size:
            page = self._get_page(pageno, False, **kwargs)
            vals.append(page.load(pageoff, size=size, endness=endness, page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs))

        else:
            max_pageno = (1 << self.state.arch.bits) // self.page_size
            bytes_done = 0
            while bytes_done < size:
                page = self._get_page(pageno, False, **kwargs)
                sub_size = min(self.page_size-pageoff, size-bytes_done)
                vals.append(page.load(pageoff, size=sub_size, endness=endness, page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs))

                bytes_done += sub_size
                pageno = (pageno + 1) % max_pageno
                pageoff = 0

        out = self.PAGE_TYPE._compose_objects(vals, size, endness, memory=self, **kwargs)
        l.debug("%s.load(%#x, %d, %s) = %s", self.id, addr, size, endness, out)
        return out

    def store(self, addr: int, data, size: int=None, endness=None, **kwargs):
        if endness is None:
            endness = self.endness

        if type(size) is not int:
            raise TypeError("Need size to be resolved to an int by this point")

        if type(addr) is not int:
            raise TypeError("Need addr to be resolved to an int by this point")

        # l.debug("%s.store(%#x, %s, %s)", self.id, addr, data, endness)

        pageno, pageoff = self._divide_addr(addr)
        sub_gen = self.PAGE_TYPE._decompose_objects(addr, data, endness, memory=self, **kwargs)
        next(sub_gen)

        # fasttrack basic case
        if pageoff + size <= self.page_size:
            written_size = 0
            while written_size < size:
                sub_data, sub_data_size = sub_gen.send(size - written_size)
                page = self._get_page(pageno, True, **kwargs)
                sub_data_size = min(sub_data_size, size - written_size)
                page.store(pageoff + written_size, sub_data, size=sub_data_size, endness=endness,
                           page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs)
                written_size += sub_data_size
            sub_gen.close()
            return

        max_pageno = (1 << self.state.arch.bits) // self.page_size
        bytes_done = 0
        while bytes_done < size:
            # if we really want we could add an optimization where writing an entire page creates a new page object
            # instead of overwriting the old one. would have to be careful about maintaining the contract for _get_page
            page = self._get_page(pageno, True, **kwargs)
            sub_size = min(self.page_size-pageoff, size-bytes_done)
            written_size = 0

            while written_size < sub_size:
                sub_data, sub_data_size = sub_gen.send(sub_size)
                sub_data_size = min(sub_data_size, sub_size - written_size)
                page.store(pageoff + written_size, sub_data, size=sub_data_size, endness=endness,
                           page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs)
                written_size += sub_data_size

            bytes_done += sub_size
            pageno = (pageno + 1) % max_pageno
            pageoff = 0

        sub_gen.close()

    def erase(self, addr, size=None, **kwargs) -> None:
        if type(size) is not int:
            raise TypeError("Need size to be resolved to an int by this point")

        if type(addr) is not int:
            raise TypeError("Need addr to be resolved to an int by this point")

        pageno, pageoff = self._divide_addr(addr)
        max_pageno = (1 << self.state.arch.bits) // self.page_size
        bytes_done = 0
        while bytes_done < size:
            page = self._get_page(pageno, True, **kwargs)
            sub_size = min(self.page_size - pageoff, size - bytes_done)
            page.erase(pageoff, sub_size, memory=self, **kwargs)
            bytes_done += sub_size
            pageno = (pageno + 1) % max_pageno
            pageoff = 0

    def merge(self, others: Iterable['PagedMemoryMixin'], merge_conditions, common_ancestor=None) -> bool:
        changed_pages_and_offsets: Dict[int,Optional[Set[int]]] = {}
        for o in others:
            for changed_page, changed_offsets in self.changed_pages(o).items():
                if changed_offsets is None:
                    changed_pages_and_offsets[changed_page] = None
                else:
                    # changed_offsets is a set of offsets (ints)
                    if changed_page not in changed_pages_and_offsets:
                        # update our dict
                        changed_pages_and_offsets[changed_page] = changed_offsets
                    else:
                        # changed_page in our dict
                        if changed_pages_and_offsets[changed_page] is None:
                            # in at least one `other` memory can we not determine the changed offsets
                            # do nothing
                            pass
                        else:
                            # union changed_offsets with known ones
                            changed_pages_and_offsets[changed_page] = \
                                changed_pages_and_offsets[changed_page].union(changed_offsets)

        if merge_conditions is None:
            merge_conditions = [None] * (len(list(others)) + 1)

        merged_bytes = set()
        for page_no in sorted(changed_pages_and_offsets.keys()):
            l.debug("... on page %x", page_no)

            page = self._get_page(page_no, True)
            other_pages = [ ]

            for o in others:
                if page_no in o._pages:
                    other_pages.append(o._get_page(page_no, False))

            page_addr = page_no * self.page_size
            changed_offsets = changed_pages_and_offsets[page_no]
            merged_offsets = page.merge(other_pages, merge_conditions, page_addr=page_addr, memory=self,
                                        changed_offsets=changed_offsets)
            for off in merged_offsets:
                merged_bytes.add(page_addr + off)

        return bool(merged_bytes)

    def permissions(self, addr, permissions=None, **kwargs):
        if type(addr) is not int:
            raise TypeError("addr must be an int in paged memory")
        pageno, _ = self._divide_addr(addr)
        try:
            page = self._get_page(pageno, permissions is not None, allow_default=False, **kwargs)
        except SimMemoryError as e:
            raise SimMemoryError("%#x is not mapped" % addr) from e

        if type(permissions) is int:
            permissions = self.state.solver.BVV(permissions, 3)

        result = page.permissions
        if permissions is not None:
            page.permissions = permissions
        return result

    def map_region(self, addr, length, permissions, init_zero=False, **kwargs):
        if type(addr) is not int:
            raise TypeError("addr must be an int in paged memory")
        pageno, pageoff = self._divide_addr(addr)
        size_done = 0
        max_pageno = (1 << self.state.arch.bits) // self.page_size

        while size_done < length:
            self._map_page(pageno, permissions, init_zero=init_zero, **kwargs)
            size_done += self.page_size - pageoff
            pageoff = 0
            pageno = (pageno + 1) % max_pageno

    def unmap_region(self, addr, length, **kwargs):
        if type(addr) is not int:
            raise TypeError("addr must be an int in paged memory")

        pageno, pageoff = self._divide_addr(addr)
        size_done = 0
        max_pageno = (1 << self.state.arch.bits) // self.page_size

        while size_done < length:
            self._unmap_page(pageno, **kwargs)
            size_done += self.page_size - pageoff
            pageoff = 0
            pageno = (pageno + 1) % max_pageno

    def _map_page(self, pageno, permissions, init_zero=False, **kwargs):
        # the logical flow of this shit is so confusing. this should be more compact and efficient than the old
        # simmemory model but it will require a bit of writing to solidify the contracts and allow someone
        # reading this function to actually believe that it is correctly implemented
        try:
            self._get_page(pageno, False, allow_default=False, **kwargs)
        except SimMemoryError:
            pass  # good, expected error
        else:
            raise SimMemoryError("Page is already mapped", pageno * self.page_size)

        page = self._initialize_default_page(pageno, permissions=permissions, **kwargs)
        self._pages[pageno] = page
        if init_zero:
            page.store(0, None, size=self.page_size, endness='Iend_BE', page_addr=pageno*self.page_size, memory=self,
                       **kwargs)

    def _unmap_page(self, pageno, **kwargs):  # pylint:disable=unused-argument
        try:
            if self._pages[pageno] is not None:
                self._pages[pageno].release_shared()
                self._pages[pageno] = None
        except KeyError:
            pass

    def __contains__(self, addr):
        pageno, _ = self._divide_addr(addr)
        try:
            self._get_page(pageno, False, allow_default=False)
        except SimMemoryError:
            return False
        else:
            return True

    def _load_to_memoryview(self, addr, size, with_bitmap):
        result = self.load(addr, size, endness='Iend_BE')
        if result.op == 'BVV':
            if with_bitmap:
                return memoryview(result.args[0].to_bytes(size, 'big')), memoryview(bytes(size))
            else:
                return memoryview(result.args[0].to_bytes(size, 'big'))
        elif result.op == 'Concat':
            bytes_out = bytearray(size)
            bitmap_out = bytearray(size)
            bit_idx = 0
            byte_width = self.state.arch.byte_width
            for element in result.args:
                byte_idx = bit_idx // byte_width
                byte_size = len(element) // byte_width

                if len(element) + bit_idx < (byte_idx + 1) * byte_width:
                    # the current element is not long enough to reach the next byte
                    # it is impossible to have multiple concrete BVV objects straddling byte boundaries
                    bit_idx += len(element)
                    if not with_bitmap:
                        return memoryview(bytes(bytes_out))[:byte_idx]
                    bitmap_out[byte_idx] = 1
                    continue

                if bit_idx % byte_width != 0:
                    # if the current element has at least byte_width bits, the top `hi_chop` bits should be removed
                    hi_chop = byte_width - (bit_idx % byte_width)
                else:
                    hi_chop = 0

                if (bit_idx + len(element)) % byte_width != 0:
                    # if the current element does not have enough bits to extend to the next byte boundary, the
                    # bottom `lo_chop` bits should be removed
                    lo_chop = (bit_idx + len(element)) % byte_width
                else:
                    lo_chop = 0

                if hi_chop + lo_chop == len(element):
                    # the entire element will be removed
                    bit_idx += len(element)
                    if not with_bitmap:
                        return memoryview(bytes(bytes_out))[:byte_idx]
                    bitmap_out[byte_idx] = 1
                    continue

                if hi_chop:
                    bitmap_out[byte_idx] = 1
                    byte_idx += 1

                if element.op == 'BVV':
                    chopped = element
                    if hi_chop:
                        chopped = chopped[len(chopped) - 1 - hi_chop:0]
                    if lo_chop:
                        chopped = chopped[:lo_chop]

                    if len(chopped) > 0:
                        bytes_out[byte_idx:byte_idx + byte_size] = chopped.args[0].to_bytes(byte_size, 'big')
                else:
                    if not with_bitmap:
                        return memoryview(bytes(bytes_out))[:byte_idx]
                    for byte_i in range(byte_idx, byte_idx + byte_size):
                        bitmap_out[byte_i] = 1

                bit_idx += len(element)
                if bit_idx % byte_width != 0:
                    if not with_bitmap:
                        return memoryview(bytes(bytes_out))[:bit_idx // byte_width]
                    bitmap_out[bit_idx // byte_width] = 1
            if with_bitmap:
                return memoryview(bytes(bytes_out)), memoryview(bytes(bitmap_out))
            else:
                return memoryview(bytes(bytes_out))
        else:
            if with_bitmap:
                return memoryview(bytes(size)), memoryview(b'\x01' * size)
            else:
                return memoryview(b'')

    def concrete_load(self, addr, size, writing=False, with_bitmap=False, **kwargs):
        pageno, offset = self._divide_addr(addr)
        subsize = min(size, self.page_size - offset)
        try:
            page = self._get_page(pageno, writing, **kwargs)
        except SimMemoryError:
            if with_bitmap:
                return memoryview(b''), memoryview(b'')
            else:
                return memoryview(b'')

        if not page.SUPPORTS_CONCRETE_LOAD:
            # the page does not support concrete_load
            return self._load_to_memoryview(addr, size, with_bitmap)

        data, bitmap = page.concrete_load(offset, subsize, **kwargs)
        if with_bitmap:
            return data, bitmap

        # everything from here on out has exactly one goal: to maximize the amount of concrete data
        # we can return (up to the limit!)
        for i, byte in enumerate(bitmap):
            if byte != 0:
                break
        else:
            i = len(bitmap)

        if i != subsize:
            return data[:i]

        size -= subsize

        physically_adjacent = True
        while size:
            offset = 0
            max_pageno = (1 << self.state.arch.bits) // self.page_size
            pageno = (pageno + 1) % max_pageno
            subsize = min(size, self.page_size)

            try:
                page = self._get_page(pageno, writing, **kwargs)
                concrete_load = page.concrete_load
            except (SimMemoryError, AttributeError):
                break
            else:

                newdata, bitmap = concrete_load(offset, subsize, **kwargs)
                for i, byte in enumerate(bitmap):
                    if byte != 0:
                        break
                else:
                    i = len(bitmap)

                # magic: check if the memory regions are physically adjacent
                if physically_adjacent and ffi.cast(ffi.BVoidP, ffi.from_buffer(data)) + len(data) == ffi.cast(ffi.BVoidP, ffi.from_buffer(newdata)):
                    # magic: generate a new memoryview which contains the two physically adjacent regions
                    obj = data.obj
                    if obj is None:
                        # XXX HACK FOR PYPY
                        obj = page.concrete_data.obj
                    data_offset = ffi.cast(ffi.BVoidP, ffi.from_buffer(data)) - ffi.cast(ffi.BVoidP,
                                                                                         ffi.from_buffer(obj))
                    data = memoryview(obj)[data_offset:data_offset + len(data) + i]
                else:
                    # they are not adjacent - create a new bytearray to hold data
                    physically_adjacent = False
                    bytes_out = bytearray(data) + bytearray(newdata[:i])
                    data = memoryview(bytes_out)

                if i != subsize:
                    break

                size -= subsize

        return data

    def changed_bytes(self, other) -> Set[int]:
        my_pages = set(self._pages)
        other_pages = set(other._pages)
        intersection = my_pages.intersection(other_pages)
        difference = my_pages.difference(other_pages)

        changes = set()
        for pageno in difference:
            changes.update(range(pageno * self.page_size, (pageno + 1) * self.page_size))

        for pageno in intersection:
            my_page = self._pages[pageno]
            other_page = other._pages[pageno]

            if (my_page is None) ^ (other_page is None):
                changes.update(range(pageno * self.page_size, (pageno + 1) * self.page_size))
            elif my_page is None:
                pass
            else:
                changes.update(my_page.changed_bytes(other_page, page_addr=pageno * self.page_size))

        return changes

    def changed_pages(self, other) -> Dict[int,Optional[Set[int]]]:
        my_pages = set(self._pages)
        other_pages = set(other._pages)
        intersection = my_pages.intersection(other_pages)
        difference = my_pages.difference(other_pages)

        changes: Dict[int,Optional[Set[int]]] = dict((d, None) for d in difference)

        for pageno in intersection:
            my_page = self._pages[pageno]
            other_page = other._pages[pageno]

            if (my_page is None) ^ (other_page is None):
                changes[pageno] = None
            elif my_page is None:
                pass
            else:
                changed_offsets = my_page.changed_bytes(other_page, page_addr=pageno * self.page_size)
                if changed_offsets:
                    changes[pageno] = changed_offsets

        return changes

    def _replace_all(self, addrs: Iterable[int], old: claripy.ast.BV, new: claripy.ast.BV):

        page_offsets: Dict[Set[int]] = defaultdict(set)
        for addr in addrs:
            page_no, page_offset = self._divide_addr(addr)
            page_offsets[page_no].add(page_offset)

        for page_no, offsets in page_offsets.items():
            page = self._pages[page_no]
            page = page.acquire_unique()
            self._pages[page_no] = page
            page.replace_all_with_offsets(offsets, old, new, memory=self)

    def copy_contents(self, dst, src, size, condition=None, **kwargs):
        data = self.load(src, size, **kwargs)
        self.store(dst, data, size, **kwargs)

    def flush_pages(self, white_list):
        """
        Flush all pages not included in the `white_list` by removing their pages. Note, this will not wipe them
        from memory if they were backed by a memory_backer, it will simply reset them to their initial state.
        Returns the list of pages that were cleared consisting of `(addr, length)` tuples.
        :param white_list: white list of regions in the form of (start, end) to exclude from the flush
        :return: a list of memory page ranges that were flushed
        :rtype: list
        """
        white_list_page_number = []

        for addr in white_list:
            for page_addr in range(addr[0], addr[1], self.page_size):
                pageno, _ = self.state.memory._divide_addr(page_addr)
                white_list_page_number.append(pageno)

        new_page_dict = {}
        flushed = []

        # cycle over all the keys ( the page number )
        for pageno, page in self._pages.items():
            if pageno in white_list_page_number:
                #l.warning("Page " + str(pageno) + " not flushed!")
                new_page_dict[pageno] = page
            else:
                #l.warning("Page " + str(pageno) + " flushed!")
                flushed.append((pageno, self.page_size))
        self._pages = new_page_dict
        return flushed


class LabeledPagesMixin(PagedMemoryMixin):
    def load_with_labels(self, addr: int, size: int=None, endness=None, **kwargs) -> Tuple[claripy.ast.Base,Tuple[Tuple[int,int,int,Any]]]:
        if endness is None:
            endness = self.endness

        if type(size) is not int:
            raise TypeError("Need size to be resolved to an int by this point")

        if type(addr) is not int:
            raise TypeError("Need addr to be resolved to an int by this point")

        pageno, pageoff = self._divide_addr(addr)
        vals = []

        # fasttrack basic case
        if pageoff + size <= self.page_size:
            page = self._get_page(pageno, False, **kwargs)
            vals.append(page.load(pageoff, size=size, endness=endness, page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs))

        else:
            max_pageno = (1 << self.state.arch.bits) // self.page_size
            bytes_done = 0
            while bytes_done < size:
                page = self._get_page(pageno, False, **kwargs)
                sub_size = min(self.page_size-pageoff, size-bytes_done)
                vals.append(page.load(pageoff, size=sub_size, endness=endness, page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs))

                bytes_done += sub_size
                pageno = (pageno + 1) % max_pageno
                pageoff = 0

        labels = [ ]
        out = self.PAGE_TYPE._compose_objects(vals, size, endness, memory=self, labels=labels, **kwargs)
        l.debug("%s.load_with_labels(%#x, %d, %s) = %s", self.id, addr, size, endness, out)
        return out, tuple(labels)


class ListPagesMixin(PagedMemoryMixin):
    PAGE_TYPE = ListPage


class MVListPagesMixin(PagedMemoryMixin):
    PAGE_TYPE = MVListPage

    def __init__(self, *args, skip_missing_values_during_merging=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.skip_missing_values_during_merging = skip_missing_values_during_merging

    @MemoryMixin.memo
    def copy(self, memo) -> 'MVListPagesMixin':
        r = super().copy(memo)
        r.skip_missing_values_during_merging = self.skip_missing_values_during_merging
        return r


class ListPagesWithLabelsMixin(
    LabeledPagesMixin,
    ListPagesMixin,
):
    pass


class MVListPagesWithLabelsMixin(
    LabeledPagesMixin,
    MVListPagesMixin,
):
    pass


class UltraPagesMixin(PagedMemoryMixin):
    PAGE_TYPE = UltraPage



import pyvex

from ..light.resilience import VEXResilienceMixin, raiseme
from ..claripy.datalayer import ClaripyDataMixin, symbol, value
from angr import sim_options as o

class HeavyResilienceMixin(VEXResilienceMixin, ClaripyDataMixin):
    @staticmethod
    def __make_default(ty, symbolic, name):
        if symbolic:
            return symbol(ty, name)
        else:
            return value(ty, 0)

    def _check_unsupported_ccall(self, func_name, retty, args, **kwargs):
        if o.BYPASS_UNSUPPORTED_IRCCALL not in self.state.options:
            return super()._check_unsupported_ccall(func_name, retty, args, **kwargs)
        self.state.history.add_event('resilience', resilience_type='ccall', callee=func_name, message='unsupported ccall')
        return self.__make_default(retty, o.UNSUPPORTED_BYPASS_ZERO_DEFAULT not in self.state.options, 'unsupported_' + func_name)

    def _check_errored_ccall(self, func_name, ty, args, **kwargs):
        if o.BYPASS_ERRORED_IRCCALL not in self.state.options:
            return super()._check_errored_ccall(func_name, ty, args, **kwargs)
        self.state.history.add_event('resilience', resilience_type='ccall', callee=func_name, message='ccall raised SimCCallError')
        return self.__make_default(ty, True, 'errored_' + func_name)

    def _check_unsupported_dirty(self, func_name, ty, args, **kwargs):
        if o.BYPASS_UNSUPPORTED_IRDIRTY not in self.state.options:
            return super()._check_unsupported_dirty(func_name, ty, args, **kwargs)
        if ty is None:
            return None
        return self.__make_default(ty, o.UNSUPPORTED_BYPASS_ZERO_DEFAULT not in self.state.options, 'unsupported_' + func_name)

    def _check_unsupported_op(self, op, args):
        ty = pyvex.get_op_retty(op)
        if o.BYPASS_UNSUPPORTED_IROP not in self.state.options:
            return super()._check_unsupported_op(op, args)
        self.state.history.add_event('resilience', resilience_type='irop', op=op, message='unsupported IROp')
        return self.__make_default(ty, o.UNSUPPORTED_BYPASS_ZERO_DEFAULT not in self.state.options, 'unsupported_' + op)

    def _check_errored_op(self, op, args):
        ty = pyvex.get_op_retty(op)
        if o.BYPASS_ERRORED_IROP not in self.state.options:
            return super()._check_errored_op(op, args)
        self.state.history.add_event('resilience', resilience_type='irop', op=op, message='unsupported IROp')
        return self.__make_default(ty, True, 'errored_' + op)

    def _check_zero_division(self, op, args):
        if getattr(self.state, 'mode', None) == 'static' and len(args) == 2 and (args[1] == 0).is_true():
            # Monkeypatch the dividend to another value instead of 0
            args = list(args)
            ty = pyvex.expr.op_arg_types(op)[1][1]
            args[1] = value(ty, 1)
            return self._perform_vex_expr_Op(op, args)
        res = super()._check_zero_division(op, args)
        if res is not raiseme:
            return res
        return self._check_errored_op(op, args)

    def _check_errored_stmt(self, stmt):
        if o.BYPASS_ERRORED_IRSTMT not in self.state.options:
            return super()._check_errored_stmt(stmt)
        self.state.history.add_event(
                'resilience',
                resilience_type='irstmt',
                stmt=type(stmt).__name__,
                message='errored IRStmt')
        return None



from typing import Optional
import logging

from . import MemoryMixin
from ...errors import SimMemoryLimitError, SimMemoryError, SimUnsatError

l = logging.getLogger(__name__)


class SizeNormalizationMixin(MemoryMixin):
    """
    Provides basic services related to normalizing sizes. After this mixin, sizes will always be a plain int.
    Assumes that the data is a BV.

    - load will throw a TypeError if no size is provided
    - store will default to len(data)//byte_width if no size is provided
    """
    def load(self, addr, size=None, **kwargs):
        if size is None:
            raise TypeError("Must provide size to load")
        elif type(size) is int:
            out_size = size
        elif getattr(size, 'op', None) == 'BVV':
            out_size = size.args[0]
        else:
            raise Exception("Size must be concretely resolved by this point in the memory stack")

        return super().load(addr, size=out_size, **kwargs)

    def store(self, addr, data, size=None, **kwargs):
        max_size = len(data) // self.state.arch.byte_width
        if size is None:
            out_size = max_size
        elif type(size) is int:
            out_size = size
        elif getattr(size, 'op', None) == 'BVV':
            out_size = size.args[0]
        else:
            raise Exception("Size must be concretely resolved by this point in the memory stack")

        if out_size > max_size:
            raise SimMemoryError("Not enough data for store")
        if out_size == 0:
            # skip zero-sized stores
            return

        super().store(addr, data, size=out_size, **kwargs)


class SizeConcretizationMixin(MemoryMixin):
    """
    This mixin allows memory to process symbolic sizes. It will not touch any sizes which are not ASTs with non-BVV ops.
    Assumes that the data is a BV.

    - symbolic load sizes will be concretized as their maximum and a warning will be logged
    - symbolic store sizes will be dispatched as several conditional stores with concrete sizes
    """
    def __init__(self, concretize_symbolic_write_size: bool=False, max_concretize_count: Optional[int]=256,
                 max_symbolic_size: int=0x400000, raise_memory_limit_error: bool=False, size_limit: int=257, **kwargs):
        super().__init__(**kwargs)
        self._concretize_symbolic_write_size = concretize_symbolic_write_size  # in place of the state option CONCRETIZE_SYMBOLIC_WRITE_SIZES
        self._max_concretize_count = max_concretize_count
        self._max_symbolic_size = max_symbolic_size
        self._raise_memory_limit_error = raise_memory_limit_error
        self._size_limit = size_limit

    def copy(self, memo):
        o = super().copy(memo)

        o._concretize_symbolic_write_size = self._concretize_symbolic_write_size
        o._max_concretize_count = self._max_concretize_count
        o._max_symbolic_size = self._max_symbolic_size
        o._raise_memory_limit_error = self._raise_memory_limit_error
        o._size_limit = self._size_limit

        return o

    def load(self, addr, size=None, **kwargs):
        if getattr(size, 'op', 'BVV') == 'BVV':
            return super().load(addr, size=size, **kwargs)

        l.warning("Loading symbolic size via max. be careful.")
        out_size = self.state.solver.max(size)
        return super().load(addr, size=out_size, **kwargs)

    def store(self, addr, data, size=None, condition=None, **kwargs):
        if getattr(size, 'op', 'BVV') == 'BVV':
            super().store(addr, data, size=size, condition=condition, **kwargs)
            return

        max_size = len(data) // self.state.arch.byte_width
        try:
            if self._raise_memory_limit_error:
                conc_sizes = list(self.state.solver.eval_upto(
                    size,
                    self._size_limit,
                    extra_constraints=(size <= max_size,)
                    )
                )
                if len(conc_sizes) == self._size_limit:
                    raise SimMemoryLimitError("Extremely unconstrained store size")
            else:
                conc_sizes = list(self.state.solver.eval_upto(
                    size,
                    self._max_concretize_count,
                    extra_constraints=(size <= max_size,)
                    )
                )
        except SimUnsatError:
            # size has to be greater than max_size
            raise SimMemoryError("Not enough data for store")

        # filter out all concrete sizes that are greater than max_size
        # Note that the VSA solver (used in static mode) cannot precisely handle extra constraints. As a result, we may
        # get conc_sizes with values that violate the extra constraint (size <= max_size).
        conc_sizes = [ cs for cs in conc_sizes if cs <= max_size ]
        conc_sizes.sort()
        if not conc_sizes:
            raise SimMemoryError("Not enough data for store")
        if self._max_concretize_count is not None:
            conc_sizes = conc_sizes[:self._max_concretize_count]

        if size.symbolic:
            if any(cs > self._max_symbolic_size for cs in conc_sizes):
                l.warning("At least one concretized size is over the limit of %d bytes. Constrain them to the limit.",
                        self._max_symbolic_size)
            conc_sizes = [min(cs, self._max_symbolic_size) for cs in conc_sizes]

        if condition is None:
            condition = self.state.solver.true
        for conc_size in conc_sizes:
            if conc_size == 0:
                continue
            super().store(addr, data, size=conc_size, condition=condition & (size == conc_size), **kwargs)





###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_68000_BE_32_Coldfire(ArchPcode):
    name = '68000:BE:32:Coldfire'
    pcode_arch = '68000:BE:32:Coldfire'
    description = 'Motorola 32-bit Coldfire'
    bits = 32
    ip_offset = 0x50
    sp_offset = 0x3c
    bp_offset = sp_offset
    instruction_endness = Endness.BE
    register_list = [
        Register('d0', 4, 0x0),
        Register('d0w', 2, 0x2),
        Register('d0b', 1, 0x3),
        Register('d1', 4, 0x4),
        Register('d1w', 2, 0x6),
        Register('d1b', 1, 0x7),
        Register('d2', 4, 0x8),
        Register('d2w', 2, 0xa),
        Register('d2b', 1, 0xb),
        Register('d3', 4, 0xc),
        Register('d3w', 2, 0xe),
        Register('d3b', 1, 0xf),
        Register('d4', 4, 0x10),
        Register('d4w', 2, 0x12),
        Register('d4b', 1, 0x13),
        Register('d5', 4, 0x14),
        Register('d5w', 2, 0x16),
        Register('d5b', 1, 0x17),
        Register('d6', 4, 0x18),
        Register('d6w', 2, 0x1a),
        Register('d6b', 1, 0x1b),
        Register('d7', 4, 0x1c),
        Register('d7w', 2, 0x1e),
        Register('d7b', 1, 0x1f),
        Register('a0', 4, 0x20),
        Register('a0w', 2, 0x22),
        Register('a0b', 1, 0x23),
        Register('a1', 4, 0x24),
        Register('a1w', 2, 0x26),
        Register('a1b', 1, 0x27),
        Register('a2', 4, 0x28),
        Register('a2w', 2, 0x2a),
        Register('a2b', 1, 0x2b),
        Register('a3', 4, 0x2c),
        Register('a3w', 2, 0x2e),
        Register('a3b', 1, 0x2f),
        Register('a4', 4, 0x30),
        Register('a4w', 2, 0x32),
        Register('a4b', 1, 0x33),
        Register('a5', 4, 0x34),
        Register('a5w', 2, 0x36),
        Register('a5b', 1, 0x37),
        Register('a6', 4, 0x38),
        Register('a6w', 2, 0x3a),
        Register('a6b', 1, 0x3b),
        Register('sp', 4, 0x3c),
        Register('a7w', 2, 0x3e),
        Register('a7b', 1, 0x3f),
        Register('tf', 1, 0x40),
        Register('svf', 1, 0x41),
        Register('ipl', 1, 0x42),
        Register('xf', 1, 0x43),
        Register('nf', 1, 0x44),
        Register('zf', 1, 0x45),
        Register('vf', 1, 0x46),
        Register('cf', 1, 0x47),
        Register('pc', 4, 0x50, alias_names=('ip',)),
        Register('fpcr', 4, 0xb0),
        Register('fpsr', 4, 0xb4),
        Register('fpiar', 4, 0xb8),
        Register('crp', 8, 0xe0),
        Register('isp', 4, 0x100),
        Register('msp', 4, 0x104),
        Register('vbr', 4, 0x108),
        Register('cacr', 4, 0x10c),
        Register('caar', 4, 0x110),
        Register('ac0', 4, 0x114),
        Register('ac1', 4, 0x118),
        Register('usp', 4, 0x11c),
        Register('tt0', 4, 0x120),
        Register('tt1', 4, 0x124),
        Register('sfc', 4, 0x140),
        Register('dfc', 4, 0x144),
        Register('tc', 4, 0x148),
        Register('itt0', 4, 0x14c),
        Register('itt1', 4, 0x150),
        Register('dtt0', 4, 0x154),
        Register('dtt1', 4, 0x158),
        Register('mmusr', 4, 0x15c),
        Register('urp', 4, 0x160),
        Register('srp', 4, 0x164),
        Register('pcr', 4, 0x168),
        Register('cac', 4, 0x16c),
        Register('buscr', 4, 0x180),
        Register('mbb', 4, 0x184),
        Register('rambar0', 4, 0x188),
        Register('rambar1', 4, 0x18c),
        Register('sr', 2, 0x200),
        Register('acusr', 2, 0x202),
        Register('glbdenom', 4, 0x300),
        Register('movemptr', 4, 0x304),
        Register('contextreg', 4, 0x400),
        Register('macsr', 4, 0x500),
        Register('acc', 4, 0x504),
        Register('mask', 4, 0x508),
        Register('emacsr', 4, 0x600),
        Register('acc0', 4, 0x604),
        Register('acc1', 4, 0x608),
        Register('acc2', 4, 0x60c),
        Register('acc3', 4, 0x610),
        Register('accext01', 4, 0x614),
        Register('accext23', 4, 0x618),
        Register('emask', 4, 0x61c),
        Register('fp0', 12, 0x700),
        Register('fp1', 12, 0x70c),
        Register('fp2', 12, 0x718),
        Register('fp3', 12, 0x724),
        Register('fp4', 12, 0x730),
        Register('fp5', 12, 0x73c),
        Register('fp6', 12, 0x748),
        Register('fp7', 12, 0x754)
    ]

register_arch(['68000:be:32:coldfire'], 32, Endness.BE, ArchPcode_68000_BE_32_Coldfire)



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_CP1600_BE_16_default(ArchPcode):
    name = 'CP1600:BE:16:default'
    pcode_arch = 'CP1600:BE:16:default'
    description = 'General Instruments CP1600'
    bits = 16
    ip_offset = 0xe
    sp_offset = 0xc
    bp_offset = sp_offset
    instruction_endness = Endness.BE
    register_list = [
        Register('r0', 2, 0x0),
        Register('r1', 2, 0x2),
        Register('r2', 2, 0x4),
        Register('r3', 2, 0x6),
        Register('r4', 2, 0x8),
        Register('r5', 2, 0xa),
        Register('r6', 2, 0xc),
        Register('r7', 2, 0xe, alias_names=('pc', 'ip')),
        Register('i', 1, 0x10),
        Register('c', 1, 0x11),
        Register('o', 1, 0x12),
        Register('z', 1, 0x13),
        Register('s', 1, 0x14),
        Register('contextreg', 4, 0x20)
    ]

register_arch(['cp1600:be:16:default'], 16, Endness.BE, ArchPcode_CP1600_BE_16_default)
